SUBSCRIBE_CURVES = "subscribe_curves"
GET_CURVE_FILTERS = "curve_filters"

CALLBACK_TYPES = (SUBSCRIBE_CURVES, GET_CURVE_FILTERS)

class Callback:
    def __init__(self, callback_type, latest=True):
        assert callback_type in CALLBACK_TYPES, (
            f"callback_type: {callback_type} not in {CALLBACK_TYPES}"
        )
        self.callback_type = callback_type
        self.latest = latest



from energyquantified.events.responses import CurvesFiltersResponse
from .base_response import ServerResponse


class ServerResponseCurvesFilters(ServerResponse):
    DATA_FILTERS_KEY = "filters"

    @staticmethod
    def from_message(json):
        response_obj = ServerResponseCurvesFilters()
        response_obj._parse_message(json)
        return response_obj

    def _set_data(self, json):
        data_obj = json.get(self.DATA_KEY)
        if data_obj is None:
            raise ValueError(
                f"Failed parsing response from stream, "
                f"missing field '{self.DATA_KEY}'"
            )
        # Parse filters
        filters = data_obj.get(self.DATA_FILTERS_KEY)
        if filters is not None:
            filters = [self._parse_curve_filter(filter) for filter in filters]
        self.data = CurvesFiltersResponse(filters=filters)



from energyquantified.events.messages.server.base import _BaseServerMessage

class ServerMessageMessage(_BaseServerMessage):
    MESSAGE_KEY = "message"

    def __init__(self, message):
        self.message = message

    @staticmethod
    def from_message(json):
        msg = json.get(ServerMessageMessage.MESSAGE_KEY)
        if msg is None:
            raise ValueError(
                f"Failed to parse StreamMessageMessage because "
                f"field '{ServerMessageMessage.MESSAGE_KEY}' is missing"
            )
        return ServerMessageMessage(msg)



from enum import Enum

_event_lookup = {}

class EventType(Enum):
    """
    A field in event objects, describing the type of event.

     * ``CURVE_UPDATE`` – Data for a curve is created or modified
     * ``CURVE_DELETE`` – Some data for a curve is deleted
     * ``CURVE_TRUNCATE`` – All data for a curve is deleted
     * ``DISCONNECTED`` – Disconnected from the stream
     * ``TIMEOUT`` – Timeout

    Check if an EventType object is for curve-, connection- or
    timeout events:

        >>> from energyquantified.events import EventType
        >>> EventType.CURVE_UPDATE.is_curve_type()
        >>> > True
        >>> EventType.CURVE_UPDATE.is_connection_type()
        >>> > False
        >>> EventType.CURVE_UPDATE.is_timeout_type()
        >>> > False
    """

    CURVE_UPDATE = ("CURVE_UPDATE", "Curve Update", True, False, False)
    CURVE_DELETE = ("CURVE_DELETE", "Curve Delete", True, False, False)
    CURVE_TRUNCATE = ("CURVE_TRUNCATE", "Curve Truncate", True, False, False)
    DISCONNECTED = ("DISCONNECTED", "Disconnected", False, True, False)
    TIMEOUT = ("TIMEOUT", "Timeout", False, False, True)

    def __init__(
        self,
        tag,
        label,
        is_curve_type,
        is_connection_type,
        is_timeout_type
    ):
        self.tag = tag
        self.label = label
        self._is_curve_type = is_curve_type
        self._is_connection_type = is_connection_type
        self._is_timeout_type = is_timeout_type
        _event_lookup[tag.lower()] = self

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def is_valid_tag(tag):
        """
        Check whether an EventType tag exists or not.

        :param tag: The tag to look up (case-insensitive)
        :type tag: str
        :return: True if the EventType tag exists, otherwise False
        :rtype: bool
        """
        return isinstance(tag, str) and tag.lower() in _event_lookup

    @staticmethod
    def by_tag(tag):
        """
        Look up an EventType by tag.

        :param tag: The tag to look up (case-insensitive)
        :type tag: str
        :return: The EventType for this tag
        :rtype: EventType
        :raises KeyError: if no EventType exists for this tag
        """
        return _event_lookup[tag.lower()]

    def is_curve_type(self):
        """
        Check if this event type is for curve events.

        :return: True if this event type is used for curve events,\
            otherwise False.
        :rtype: bool
        """
        return self._is_curve_type

    def is_connection_type(self):
        """
        Check if this event type is for connection events.

        :return: True if this event type is used by connection events,\
            otherwise False.
        :rtype: bool
        """
        return self._is_connection_type

    def is_timeout_type(self):
        """
        Check if this event type is for timeout events.

        :return: True if this event type is used by timeout events,\
            otherwise False.
        :rtype: bool
        """
        return self._is_timeout_type



from energyquantified.metadata import CurveType
from .event_type import EventType

class _Event:
    def __init__(self, event_type):
        self._set_event_type(event_type)

    @property
    def is_curve_event(self):
        return self.event_type.is_curve_type()

    @property
    def is_connection_event(self):
        return self.event_type.is_connection_type()

    @property
    def is_timeout_event(self):
        return self.event_type.is_timeout_type()

    def _set_event_type(self, event_type):
        raise NotImplementedError

class TimeoutEvent(_Event):
    """
    Filler event used to indicate that a certain time has passed without
    any new events.
    """

    def __init__(self):
        #: See :py:class:`energyquantified.events.EventType`. Is always
        #: ``TIMEOUT`` for timeout events.
        self.event_type = None
        super().__init__(event_type=EventType.TIMEOUT)

    def _set_event_type(self, event_type):
        assert event_type.is_timeout_type(), (
            f"Cannot create TimeoutEvent with EventType={event_type}"
        )
        self.event_type = event_type

    def __str__(self):
        return (
            f"<TimeoutEvent: "
            f"event_type={self.event_type}"
            f">"
        )

    def __repr__(self) -> str:
        return self.__str__()


class CurveUpdateEvent(_Event):
    """
    Describes change in data for a :py:class:`energyquantified.metadata.Curve`,
    which may also be related to an
    :py:class:`energyquantified.metadata.Instance`.
    """

    def __init__(
        self,
        event_id,
        curve,
        event_type,
        begin = None,
        end = None,
        instance = None,
        num_values = None,
    ):
        #: See :py:class:`energyquantified.events.EventType`. Is one of
        #: ``CURVE_UPDATE``, ``CURVE_DELETE`` or ``CURVE_TRUNCATE``.
        self.event_type = None
        super().__init__(event_type=event_type)
        #: The unique identifier for this event
        self.event_id = event_id
        #: See :py:class:`energyquantified.metadata.Curve`. The curve with
        #: modified data.
        self.curve = curve
        #: See :py:class:`energyquantified.metadata.Instance`. The instace
        #: with modified data. Only provided where applicable.
        self.instance = instance
        #: The begin date (inclusive)
        self.begin = begin
        #: The end date (exclusive)
        self.end = end
        #: The number of affected values
        self.num_values = num_values

    def _set_event_type(self, event_type):
        assert event_type.is_curve_type(), (
            f"Cannot create CurveUpdateEvent with EventType={event_type}"
        )
        self.event_type = event_type

    def __str__(self):
        begin_str = self.begin.isoformat(sep=" ") if self.begin is not None else None
        end_str = self.end.isoformat(sep=" ") if self.end is not None else None
        return (
            f"<CurveUpdateEvent: "
            f"event_id={self.event_id}, "
            f"curve={self.curve}, "
            f"event_type={self.event_type}, "
            f"begin={begin_str}, "
            f"end={end_str}, "
            f"instance={self.instance}, "
            f"num_values:{self.num_values}"
            f">"
        )

    def __repr__(self):
        return self.__str__()

    def load_data(self, eq):
        """
        Load data in the range described by the event. Returns 'None' if the
        event type is either ``EventType.CURVE_TRUNCATE`` or
        ``EventType.CURVE_DELETE``.

        :param eq: Instance of the api client
        :type eq: :py:class:`energyquantified.EnergyQuantified`
        :return: The range of data the event describes
        :rtype: :py:class:`energyquantified.data.TimeSeries`,\
            :py:class:`energyquantified.data.Periodseries`,\
            :py:class:`energyquantified.data.OHLCList`
        :raises ValueError: Invalid object
        :raises APIError: If there were any network- or server-related \
            issues while loading the data
        """
        if not self.event_type == EventType.CURVE_UPDATE:
            raise ValueError(
                f"Cannot load data for curve event with EventType={self.event_type}"
            )
        # Timeseries and scenarios
        if self.curve.curve_type in (
            CurveType.TIMESERIES,
            CurveType.SCENARIO_TIMESERIES
        ):
            return eq.timeseries.load(
                self.curve,
                begin=self.begin,
                end=self.end,
            )
        # Instance
        if self.curve.curve_type == CurveType.INSTANCE:
            if self.instance is None:
                raise ValueError(
                    f"Cannot load data for event: {self}, "
                    f"instance cannot be None for "
                    f"curve type: {CurveType.INSTANCE}"
                )
            return eq.instances.get(
                self.curve,
                issued=self.instance.issued,
                tag=self.instance.tag,
            )
        # Periods
        if self.curve.curve_type == CurveType.PERIOD:
            if self.curve.frequency.is_iterable:
                raise ValueError(
                    f"Cannot load data for event: {self}, "
                    f"curve.frequency: {self.curve.frequency} "
                    f"cannot be iterable for curve type: {CurveType.PERIOD}"
                )
            # Periods
            return eq.periods.load(
                self.curve,
                begin=self.begin,
                end=self.end,
            )
        # Instance periods
        if self.curve.curve_type == CurveType.INSTANCE_PERIOD:
            if self.instance is None:
                raise ValueError(
                    f"Cannot load data for event: {self}, "
                    f"instance cannot be None for "
                    f"curve type: {CurveType.INSTANCE_PERIOD}"
                )
            if self.curve.frequency.is_iterable:
                raise ValueError(
                    f"Cannot load data for event: {self}, "
                    f"curve.frequency: {self.curve.frequency} "
                    f"cannot be iterable for "
                    f"curve type: {CurveType.INSTANCE_PERIOD}"
                )
            # InstancePeriod
            return eq.period_instances.get(
                self.curve,
                begin=self.begin,
                end=self.end,
                issued=self.instance.issued,
                tag=self.instance.tag,
            )
        # OHLC
        if self.curve.curve_type == CurveType.OHLC:
            return eq.ohlc.load(
                self.curve,
                begin=self.begin,
                end=self.end,
            )



from energyquantified.events.messages.server.base import _BaseServerMessage
from energyquantified.events import CurveUpdateEvent, EventType
from energyquantified.parser.metadata import parse_curve, parse_instance
from energyquantified.metadata import CurveType
from dateutil.parser import isoparse
from energyquantified.time import to_timezone


class ServerMessageCurveEvent(_BaseServerMessage):
    EVENT_KEY = "event"

    def __init__(self, event):
        self.event = event

    @staticmethod
    def from_message(json):
        event_json = json.get(ServerMessageCurveEvent.EVENT_KEY)
        if event_json is None:
            raise ValueError(
                f"Failed to parse StreamMessageEvent because "
                f"field '{ServerMessageCurveEvent.EVENT_KEY}' is missing"
            )
        event = _parse_curve_event(event_json)
        return ServerMessageCurveEvent(event)

def _parse_curve_event(json):
    # Curve
    curve = parse_curve(json["curve"])
    # Begin and end
    begin = json.get("begin")
    if begin is not None:
        begin = isoparse(begin)
        begin = to_timezone(begin, curve.timezone)
    end = json.get("end")
    if end is not None:
        end = isoparse(end)
        end = to_timezone(end, curve.timezone)
    # Instance
    instance = json.get("instance")
    if instance is not None:
        instance = parse_instance(instance, curve=curve)
    return CurveUpdateEvent(
        json["id"],
        curve,
        event_type=EventType.by_tag(json["event_type"]),
        begin=begin,
        end=end,
        instance=instance,
        num_values=json.get("values_changed"),
    )


from dateutil import parser
from .metadata import (
    parse_curve,
    parse_instance,
    parse_resolution,
)
from ..metadata import (
    Aggregation,
    Filter,
)
from ..data import (
    AbsoluteResult,
    AbsoluteItem,
)
from ..time import to_timezone


def parse_absolute(json):
    """
    Parse a JSON response from the API into an AbsoluteResult object with
    Curve, Resolution, delivery, Filter, Aggregation, unit and AbsoluteItem's.
    """
    # Parse the Curve
    curve = json.get("curve")
    if curve:
        curve = parse_curve(curve)
    # Parse the Resolution
    resolution = json.get("resolution")
    if resolution:
        resolution = parse_resolution(resolution)
    # Parse the delivery
    delivery = to_timezone(
            parser.isoparse(json.get("delivery")),
            tz=resolution.timezone,
    )
    # Parse the Filter
    filters = json.get("filter")
    if filters is not None:
        filters = Filter.by_tag(filters)
    # Parse the Aggregation
    aggregation = json.get("aggregation")
    if aggregation is not None:
        aggregation = Aggregation.by_tag(aggregation)
    # Parse the unit
    unit = json.get("unit")
    # Parse data
    absolute_items = _parse_items(json.get("items"), curve=curve)
    return AbsoluteResult(
        curve=curve,
        resolution=resolution,
        delivery=delivery,
        filters=filters,
        aggregation=aggregation,
        unit=unit,
        items=absolute_items,
    )

def _parse_items(json, curve):
    """
    Parse the items[] part of the json response
    """
    items = []
    for element in json:
        # Parse the instance
        instance = parse_instance(element.get("instance"), curve=curve)
        value = element.get("v")
        items.append(AbsoluteItem(instance=instance, value=value))
    return items


import pytz
import tzlocal

from ._timezone_gas import _build_europe_gas_day_tzinfo

# Look up the most used timezones in the European power markets
# from the pytz library

UTC = pytz.UTC
CET = pytz.timezone("CET")
WET = pytz.timezone("WET")
EET = pytz.timezone("EET")
TRT = pytz.timezone("Europe/Istanbul")
GAS_DAY = _build_europe_gas_day_tzinfo()


def local_tz():
    """
    Get the local timezone.

    :return: The timezone for this system.
    :rtype: TzInfo
    """
    return tzlocal.get_localzone()


def _is_valid_timezone(tz):
    """
    Check if a timezone is a valid timezone.

    :param tz: Timezone to check
    :type tz: str, pytz.tzinfo.BaseTzInfo, required
    :return: True if tz is a valid timezone, else False
    :rtype: bool
    """
    if tz is None:
        return False
    if isinstance(tz, str):
        try:
            tz = pytz.timezone(tz)
        except pytz.exceptions.UnknownTimeZoneError:
            return False
    if isinstance(tz, pytz.tzinfo.BaseTzInfo):
        return tz in (UTC, CET, WET, EET, TRT, GAS_DAY)
    return False


# Default timezone

DEFAULT_TZ = CET
LOCAL_TZ = local_tz()



from .connection_event import ConnectionEvent
from .event_type import EventType
from .event_options import CurveAttributeFilter, CurveNameFilter
from .events import CurveUpdateEvent, TimeoutEvent
from .responses import CurvesSubscribeResponse

__all__ = [
    # Types
    "EventType",
    # Options
    "CurveAttributeFilter",
    "CurveNameFilter",
    # Events
    "CurveUpdateEvent",
    "ConnectionEvent",
    "TimeoutEvent",
    # Responses
    "CurvesSubscribeResponse",
]



from energyquantified.events.responses import CurvesSubscribeResponse
from .base_response import ServerResponse


class ServerResponseCurvesSubscribe(ServerResponse):
    DATA_FILTERS_KEY = "filters"
    DATA_LAST_ID_KEY = "last_id"

    @staticmethod
    def from_message(json):
        response_obj = ServerResponseCurvesSubscribe()
        response_obj._parse_message(json)
        return response_obj

    def _set_data(self, json):
        data_obj = json.get(self.DATA_KEY)
        if data_obj is None:
            raise ValueError(
                f"Failed parsing response from stream, "
                f"missing field '{self.DATA_KEY}'"
            )
        # Last id
        last_id = data_obj.get(self.DATA_LAST_ID_KEY)
        # Filters
        filters = data_obj.get(self.DATA_FILTERS_KEY)
        if filters is not None:
            filters = [self._parse_curve_filter(filter) for filter in filters]
        self.data = CurvesSubscribeResponse(filters=filters, last_id=last_id)



from .metadata import parse_subscription
from ..user import AccountManager, Organization, User


def parse_user(json):
    """
    Parse a JSON response from the server into a User object.
    """
    name = json.get("name")
    email = json.get("email")
    organization = json.get("organization")
    if organization:
        organization = parse_organization(organization)
    subscriptions = json.get("subscriptions")
    if subscriptions:
        subscriptions = parse_subscriptions(subscriptions)
    return User(name, email, organization, subscriptions)


def parse_organization(json):
    """
    Parse a JSON response from the server into an Organization object.
    """
    name = json.get("name")
    account_manager = json.get("account_manager")
    if account_manager:
        account_manager = parse_account_manager(account_manager)
    return Organization(name, account_manager)


def parse_account_manager(json):
    """
    Parse a JSON response from the server into an AccountManager object.
    """
    name = json.get("name")
    email = json.get("email")
    return AccountManager(name, email)


def parse_subscriptions(json):
    """
    Parse a JSON response from the server into a list of Subscription objects.
    """
    return [parse_subscription(subscription) for subscription in json]
